@@ -59,7 +59,6 @@ struct system_data_t {
 
    int orientation;
 
-   bool is_2_1;   /* is running on Android OS 2.1? */
    bool paused;
 };
 
@@ -120,6 +119,8 @@ static void android_cleanup(bool uninstall_system)
       return;
    }
 
+   already_cleaned_up = true;
+
    if (uninstall_system) {
       /* I don't think android calls our atexit() stuff since we're in a shared lib
          so make sure al_uninstall_system is called */
@@ -129,8 +130,6 @@ static void android_cleanup(bool uninstall_system)
    finish_activity(_al_android_get_jnienv());
 
    (*javavm)->DetachCurrentThread(javavm);
-
-   already_cleaned_up = true;
 }
 
 static void *android_app_trampoline(ALLEGRO_THREAD *thr, void *arg)
@@ -245,10 +244,6 @@ JNI_FUNC(bool, AllegroActivity, nativeOnCreate, (JNIEnv *env, jobject obj))
    ALLEGRO_DEBUG("init time");
    _al_unix_init_time();
 
-   // Android 2.1 has a bug with glClear we have to work around
-   const char *ver = _real_al_android_get_os_version(env);
-   system_data.is_2_1 = (0 == strncmp(ver, "2.1", 3));
-
    const char *user_lib_name = al_cstr(system_data.user_lib_name);
    ALLEGRO_DEBUG("load user lib: %s", user_lib_name);
    system_data.user_lib = dlopen(user_lib_name, RTLD_LAZY|RTLD_GLOBAL);
@@ -285,7 +280,8 @@ JNI_FUNC(bool, AllegroActivity, nativeOnCreate, (JNIEnv *env, jobject obj))
 
 JNI_FUNC(void, AllegroActivity, nativeOnPause, (JNIEnv *env, jobject obj))
 {
-   (void)env; (void)obj;
+   (void)env;
+   (void)obj;
 
    ALLEGRO_DEBUG("pause activity\n");
 
@@ -293,12 +289,15 @@ JNI_FUNC(void, AllegroActivity, nativeOnPause, (JNIEnv *env, jobject obj))
 
    ALLEGRO_SYSTEM *sys = (void *)al_get_system_driver();
 
-   if(!system_data.system || !sys) {
+   if (!system_data.system || !sys) {
       ALLEGRO_DEBUG("no system driver");
       return;
    }
 
-   // ASSERT(sys != NULL);
+   if (!_al_vector_size(&sys->displays)) {
+      ALLEGRO_DEBUG("no display, not sending SWITCH_OUT event");
+      return;
+   }
 
    ALLEGRO_DISPLAY *display = *(ALLEGRO_DISPLAY**)_al_vector_ref(&sys->displays, 0);
 
@@ -312,10 +311,6 @@ JNI_FUNC(void, AllegroActivity, nativeOnPause, (JNIEnv *env, jobject obj))
          _al_event_source_emit_event(&display->es, &event);
       }
       _al_event_source_unlock(&display->es);
-
-      ALLEGRO_DISPLAY_ANDROID *d = (ALLEGRO_DISPLAY_ANDROID*)display;
-      if (d->created)
-         _al_android_destroy_surface(env, d, false);
    }
 }
 
@@ -362,6 +357,7 @@ JNI_FUNC(void, AllegroActivity, nativeOnResume, (JNIEnv *env, jobject obj))
    }
 }
 
+/* NOTE: don't put any ALLEGRO_DEBUG in here! */
 JNI_FUNC(void, AllegroActivity, nativeOnDestroy, (JNIEnv *env, jobject obj))
 {
    (void)obj;
@@ -381,20 +377,17 @@ JNI_FUNC(void, AllegroActivity, nativeOnDestroy, (JNIEnv *env, jobject obj))
       "getMainReturned",
       "()Z"
    );
+
    if (!main_returned) {
-      android_cleanup(false);
-      return;
+      exit(0);
    }
 
-   ALLEGRO_DEBUG("destroy activity");
    if(!system_data.user_lib) {
-      ALLEGRO_DEBUG("user lib not loaded.");
       return;
    }
 
    system_data.user_main = NULL;
    if(dlclose(system_data.user_lib) != 0) {
-      ALLEGRO_ERROR("failed to unload user lib: %s", dlerror());
       return;
    }
 
@@ -465,11 +458,10 @@ JNI_FUNC(void, AllegroActivity, nativeSendJoystickConfigurationEvent, (JNIEnv *e
    _al_event_source_unlock(es);
 }
 
+/* NOTE: don't put any ALLEGRO_DEBUG in here! */
 static void finish_activity(JNIEnv *env)
 {
-   ALLEGRO_DEBUG("pre post");
    _jni_callVoidMethod(env, system_data.activity_object, "postFinish");
-   ALLEGRO_DEBUG("post post");
 }
 
 static ALLEGRO_SYSTEM *android_initialize(int flags)
@@ -541,6 +533,7 @@ ALLEGRO_SYSTEM_INTERFACE *_al_system_android_interface()
    android_vt = al_malloc(sizeof *android_vt);
    memset(android_vt, 0, sizeof *android_vt);
 
+   android_vt->id = ALLEGRO_SYSTEM_ID_ANDROID;
    android_vt->initialize = android_initialize;
    android_vt->get_display_driver = _al_get_android_display_driver;
    android_vt->get_keyboard_driver = _al_get_android_keyboard_driver;
@@ -605,11 +598,6 @@ const char *al_android_get_os_version(void)
    return _real_al_android_get_os_version(_al_android_get_jnienv());
 }
 
-bool _al_android_is_os_2_1(void)
-{
-   return system_data.is_2_1;
-}
-
 void _al_android_thread_created(void)
 {
    JNIEnv *env;
@@ -651,4 +639,18 @@ void _al_register_system_interfaces(void)
    /* TODO: add the non native activity driver */
 }
 
+/* Function: al_android_get_jni_env
+ */
+JNIEnv *al_android_get_jni_env(void)
+{
+   return _al_android_get_jnienv();
+}
+
+/* Function: al_android_get_activity
+ */
+jobject al_android_get_activity(void)
+{
+   return _al_android_activity_object();
+}
+
 /* vim: set sts=3 sw=3 et: */
