@@ -48,6 +48,16 @@ void al_set_new_display_option(int option, int value, int importance)
 }
 
 
+int _al_get_suggested_display_option(ALLEGRO_DISPLAY *d,
+   int option, int default_value)
+{
+   ALLEGRO_EXTRA_DISPLAY_SETTINGS *s = &d->extra_settings;
+   uint64_t flags = s->required | s->suggested;
+   if (flags & ((uint64_t)1 << option))
+      return s->settings[option];
+   return default_value;
+}
+
 
 /* Function: al_get_new_display_option
  */
