@@ -109,7 +109,10 @@ static ALLEGRO_PATH *early_get_exename_path(void)
 #elif defined(ALLEGRO_ANDROID)
    return _al_android_get_path(ALLEGRO_EXENAME_PATH);
 #elif defined(ALLEGRO_SDL)
-   return al_create_path_for_directory(SDL_GetBasePath());
+   char* p = SDL_GetBasePath();
+   ALLEGRO_PATH *path = al_create_path_for_directory(p);
+   SDL_free(p);
+   return path;
 #else
    #error early_get_exename_path not implemented
 #endif
@@ -142,6 +145,12 @@ static void read_allegro_cfg(void)
          al_merge_config_into(sys_config, temp);
          al_destroy_config(temp);
       }
+      al_set_path_filename(path, ".allegro5rc");
+      temp = al_load_config_file(al_path_cstr(path, '/'));
+      if (temp) {
+         al_merge_config_into(sys_config, temp);
+         al_destroy_config(temp);
+      }
       al_destroy_path(path);
    }
 #endif
@@ -222,6 +231,7 @@ bool al_install_system(int version, int (*atexit_ptr)(void (*)(void)))
       return false;
 
    _al_tls_init_once();
+   _al_reinitialize_tls_values();
 
    _al_vector_init(&_al_system_interfaces, sizeof(ALLEGRO_SYSTEM_INTERFACE *));
 
@@ -282,7 +292,9 @@ bool al_install_system(int version, int (*atexit_ptr)(void (*)(void)))
       active_sysdrv->vt->heartbeat_init();
 
    if (atexit_ptr && atexit_virgin) {
+#ifndef ALLEGRO_ANDROID
       atexit_ptr(al_uninstall_system);
+#endif
       atexit_virgin = false;
    }
 
@@ -343,6 +355,13 @@ ALLEGRO_SYSTEM *al_get_system_driver(void)
    return active_sysdrv;
 }
 
+/* Function: al_get_system_id
+ */
+ALLEGRO_SYSTEM_ID al_get_system_id(void)
+{
+   ASSERT(active_sysdrv);
+   return active_sysdrv->vt->id;
+}
 
 /* Function: al_get_system_config
  */
